from osm2gmns.networkclass.macronet import Node, Network
import osm2gmns.settings as og_settings
from shapely import geometry
import csv
import os
import sys


def _designateComplexIntersectionsFromIntFile(network, int_file, int_buffer):
    if not os.path.exists(int_file):
        sys.exit(f'ERROR: int_file {int_file} does not exist')

    fin = open(int_file, 'r')
    reader = csv.DictReader(fin)

    for field in ['x_coord', 'y_coord']:
        if field not in reader.fieldnames:
            sys.exit(f'ERROR: required field ({field}) does not exist in the int_file')

    max_intersection_id = network.max_intersection_id
    for int_info in reader:
        int_center = geometry.Point(float(int_info['x_coord']), float(int_info['y_coord']))
        int_center_xy = network.GT.geo_from_latlon(int_center)
        if 'int_buffer' in reader.fieldnames:
            buffer_ = int_info['int_buffer']
            buffer_ = float(buffer_) if buffer_ else int_buffer
        else:
            buffer_ = int_buffer

        intersection_nodes = [node for _, node in network.node_dict.items() if node.intersection_id is None and int_center_xy.distance(node.geometry_xy) <= buffer_]

        if len(intersection_nodes) < 2:
            continue

        for node in intersection_nodes:
            node.intersection_id = max_intersection_id
        max_intersection_id += 1

    network.max_intersection_id = max_intersection_id



def _autoidentifyComplexIntersections(network, int_buffer):
    """
    Identify nodes that belongs to one intersection in real life. Nodes that
    belong to one intersection will be assigned with the same intersection_id.
    Only signalized nodes will be checked.

    The reason why only signalized nodes will be checked is that we use a
    distance-based rule here, and there are many short links in osm. If all
    nodes are checked, some short links will be indentified as intersections
    by mistake.

    Parameters
    ----------
    network : Network
        Network instance
    int_buffer : float
        A threshold to check if two connected nodes belong to one intersection.
        If the length of a link that connects two nodes is shorter than int_buffer,
        these two nodes come from one intersection.

    Returns
    -------
    None
    """

    group_list = []
    group_status = []
    for _,link in network.link_dict.items():
        if link.length > int_buffer: continue
        if not (link.from_node.intersection_id is None and link.to_node.intersection_id is None): continue
        if not (link.from_node.ctrl_type == 'signal' and link.to_node.ctrl_type == 'signal'): continue
        group_list.append({link.from_node, link.to_node})
        group_status.append(1)

    number_of_valid_groups = sum(group_status)
    while True:
        for group_no1,group1 in enumerate(group_list):
            if group_status[group_no1] == 0: continue
            for group_no2,group2 in enumerate(group_list):
                if group_status[group_no2] == 0: continue
                if group_no1 == group_no2: continue
                if len(group1.intersection(group2)) > 0:
                    group1.update(group2)
                    group_status[group_no2] = 0

        new_number_of_valid_groups = sum(group_status)
        if number_of_valid_groups == new_number_of_valid_groups:
            break
        else:
            number_of_valid_groups = new_number_of_valid_groups

    max_intersection_id = network.max_intersection_id
    for group_no, group in enumerate(group_list):
        if group_status[group_no] == 0: continue
        for node in group: node.intersection_id = max_intersection_id
        max_intersection_id += 1
    network.max_intersection_id = max_intersection_id


def consolidateComplexIntersections(network, auto_identify=False, intersection_file=None, int_buffer=og_settings.default_int_buffer):
    """
    Consolidate each complex intersection that are originally represented by multiple nodes in osm into one node. Nodes
    with the same intersection_id will be consolidated into one node. intersection_id of nodes can be obtained in three ways.

    (1) set the argument auto_identify as True, then osm2gmns will automatically identify complex intersections and assign
    intersection_id for corresponding nodes.

    (2) provide an intersection file that specifies the central position (required) and buffer (optional) of each complex intersection.

    (3) user can assign intersection_id to nodes manually in network csv files (node.csv), and load the network using function loadNetFromCSV provided by osm2gmns.

    The priority of the three approaches is (3) > (2) > (1).
    Rules used in the approach (1) to identify if two nodes belong to a complex intersection: (a) ctrl_type of the two nodes must be signal;
    (b) there is a link connecting these two nodes, and the length of the link is shorter than or equal to the argument int_buffer.

    Parameters
    ----------
    network: Network
        osm2gmns Network object
    auto_identify: bool
        if automatically identify complex intersections using built-in methods in osm2gmns. nodes that belong to a complex
        intersection will be assigned with the same intersection_id
    intersection_file: str
        path of an intersction csv file that specifies complex intersections. required fields: central position of intersections
        (in the form of x_coord and y_coord); optional field: int_buffer (if not specified, the global int_buffer will be used,
        i.e., the forth arugment). For each record in the intersection_file, osm2gmns consolidates all nodes with a distance to the
        central position shorter than buffer.
    int_buffer: float
        the threshold used to check if two nodes belong to one complex intersection. the unit is meter

    Returns
    -------
    None

    """

    if intersection_file is not None:
        _designateComplexIntersectionsFromIntFile(network, intersection_file, int_buffer)

    if auto_identify:
        _autoidentifyComplexIntersections(network, int_buffer)

    if og_settings.verbose:
        print('Consolidating Complex Intersections')

    node_group_dict = {}
    node_group_ctrl_type_dict = {}
    for _, node in network.node_dict.items():
        if node.intersection_id is not None:
            if node.intersection_id in node_group_dict.keys():
                node_group_dict[node.intersection_id].append(node)
            else:
                node_group_dict[node.intersection_id] = [node]
                node_group_ctrl_type_dict[node.intersection_id] = False
            if node.ctrl_type == 'signal':
                node_group_ctrl_type_dict[node.intersection_id] = True

    removal_node_set = set()
    removal_link_set = set()
    number_of_intersections_consolidated = 0

    for intersection_id, node_group in node_group_dict.items():
        if len(node_group) < 2:
            continue

        new_node = Node(network.max_node_id)
        new_node.intersection_id = intersection_id
        if node_group_ctrl_type_dict[intersection_id]:
            new_node.ctrl_type = 'signal'
        osm_node_id_list = []
        x_coord_sum, y_coord_sum = 0.0, 0.0
        x_coord_xy_sum, y_coord_xy_sum = 0.0, 0.0

        for node in node_group:
            removal_node_set.add(node)
            osm_node_id_list.append(node.osm_node_id if node.osm_node_id is not None else 'None')
            x_coord_sum += node.geometry.x
            y_coord_sum += node.geometry.y
            x_coord_xy_sum += node.geometry_xy.x
            y_coord_xy_sum += node.geometry_xy.y

            for link in node.incoming_link_list:
                if link.from_node in node_group:
                    removal_link_set.add(link)
                else:
                    link.to_node = new_node
                    new_node.incoming_link_list.append(link)
            for link in node.outgoing_link_list:
                if link.to_node in node_group:
                    removal_link_set.add(link)
                else:
                    link.from_node = new_node
                    new_node.outgoing_link_list.append(link)

            new_node.osm_highway = node.osm_highway

        new_node.osm_node_id = '_'.join(osm_node_id_list)
        x_coord_ave = round(x_coord_sum / len(node_group), og_settings.lonlat_coord_precision)
        y_coord_ave = round(y_coord_sum / len(node_group), og_settings.lonlat_coord_precision)
        new_node.geometry = geometry.Point(x_coord_ave, y_coord_ave)
        x_coord_xy_ave = round(x_coord_xy_sum / len(node_group), og_settings.local_coord_precision)
        y_coord_xy_ave = round(y_coord_xy_sum / len(node_group), og_settings.local_coord_precision)
        new_node.geometry_xy = geometry.Point(x_coord_xy_ave, y_coord_xy_ave)

        # Keep the geometry of the links up to date by adding a small edge to join to the newly created node
        for link in new_node.incoming_link_list:
            new_coordinates = list(link.geometry.coords) + list(link.to_node.geometry.coords)
            link.geometry = geometry.LineString(new_coordinates)
        for link in new_node.outgoing_link_list:
            new_coordinates = list(link.from_node.geometry.coords) + list(link.geometry.coords)
            link.geometry = geometry.LineString(new_coordinates)

        network.node_dict[new_node.node_id] = new_node
        network.max_node_id += 1
        number_of_intersections_consolidated += 1

    for node in removal_node_set: del network.node_dict[node.node_id]
    for link in removal_link_set: del network.link_dict[link.link_id]

    if og_settings.verbose:
        print(f'    {number_of_intersections_consolidated} intersections have been consolidated')


# @author       Jiawei Lu (jiaweil9@asu.edu)
# @time         2021/11/27 12:05
# @desc         [script description]
import copy

from osm2gmns.networkclass.macronet import Network
from osm2gmns.multiresolutionnet.netgen import NetGenerator
from osm2gmns.movement.generate_movements import generateMovements, validateUserInputMovements
from osm2gmns.utils.util_geo import getLineAngle, offsetLine
import osm2gmns.settings as og_settings
from shapely import geometry
from shapely.ops import substring
import math


# todo: connector and centroid

_length_of_short_cut = 0.1
_length_of_cut = {0: 2.0, 1: 8.0, 2: 12.0, 3: 14.0, 4: 16.0, 5: 18.0, 6: 20, 7:22, 8:24}  # e.g. 2:8.0 cut 8 meters if the original macro link has 2 lanes, etc
for i_ in range(9,100): _length_of_cut[i_] = 25
_minimum_cutted_length = 2.0


def _checkMovementLinkNecessity(node_dict):
    for node_id, node in node_dict.items():
        if node.ctrl_type == 'signal': continue

        if len(node.incoming_link_list) == 1 and len(node.outgoing_link_list) >= 1:
            # one imcoming link
            ib_link = node.incoming_link_list[0]

            angle_flag = True
            for ob_link in node.outgoing_link_list:
                angle = getLineAngle(ib_link.geometry_xy, ob_link.geometry_xy)
                if angle > 0.75 * math.pi or angle < -0.75 * math.pi:
                    angle_flag = False
                    break
            if not angle_flag: continue

            ob_link_set = set()
            multiple_connection = False
            for movement in node.movement_list:
                if movement.ob_link in ob_link_set:
                    multiple_connection = True
                    break
                else:
                    ob_link_set.add(movement.ob_link)
            if multiple_connection: continue

            node.movement_link_needed = False
            ib_link.downstream_short_cut = True
            ib_link.downstream_is_target = True
            for ob_link in node.outgoing_link_list:
                ob_link.upstream_short_cut = True

        elif len(node.outgoing_link_list) == 1 and len(node.incoming_link_list) >= 1:
            # one outgoing link
            ob_link = node.outgoing_link_list[0]

            angle_flag = True
            for ib_link in node.incoming_link_list:
                angle = getLineAngle(ib_link.geometry_xy, ob_link.geometry_xy)
                if angle > 0.75 * math.pi or angle < -0.75 * math.pi:
                    angle_flag = False
                    break
            if not angle_flag: continue

            ib_link_set = set()
            multiple_connection = False
            for movement in node.movement_list:
                if movement.ib_link in ib_link_set:
                    multiple_connection = True
                    break
                else:
                    ib_link_set.add(movement.ib_link)
            if multiple_connection: continue

            node.movement_link_needed = False

            ob_link.upstream_short_cut = True
            ob_link.upstream_is_target = True
            for ib_link in node.incoming_link_list:
                ib_link.downstream_short_cut = True


def _offsetLinkGeometry(link_dict, width_of_lane, GT):
    # check if a link needs offset
    link_offset_dict = {}
    link_ids = tuple(link_dict.keys())
    for link_no_a, link_id_a in enumerate(link_ids):
        link_a = link_dict[link_id_a]
        if link_a in link_offset_dict.keys():
            continue
        geometry_xy_a_r = geometry.LineString(list(link_a.geometry_xy.coords)[::-1])
        reversed_link_found = False
        for link_id_b in link_ids[link_no_a+1:]:
            link_b = link_dict[link_id_b]
            if geometry_xy_a_r.equals_exact(link_b.geometry_xy, tolerance=0.1):
                reversed_link_found = True
                link_offset_dict[link_a] = True
                link_offset_dict[link_b] = True
                break
        if not reversed_link_found:
            link_offset_dict[link_a] = False

    # offset links
    for link, need_offset in link_offset_dict.items():
        if need_offset:
            offset_distance = (link.max_lanes / 2 + 0.5) * width_of_lane
            geometry_xy_offset = link.geometry_xy.offset_curve(distance=-1*offset_distance, join_style=2)
            if isinstance(geometry_xy_offset, geometry.MultiLineString):
                link.geometry_xy_offset = offsetLine(link.geometry_xy, offset_distance)
            else:
                link.geometry_xy_offset = geometry_xy_offset



            # offset_distance = (link.max_lanes / 2 + 0.5) * width_of_lane
            # geometry_xy_offset_ = link.geometry_xy.parallel_offset(distance=offset_distance, side='right', join_style=2)
            # if isinstance(geometry_xy_offset_, geometry.MultiLineString):
            #     link.geometry_xy_offset = offsetLine(link.geometry_xy, offset_distance)
            # else:
            #     link.geometry_xy_offset = geometry.LineString(list(geometry_xy_offset_.coords)[::-1])


            link.geometry_offset = GT.geo_to_latlon(link.geometry_xy_offset)
        else:
            link.geometry_offset = link.geometry
            link.geometry_xy_offset = link.geometry_xy

    # update breakpoints due to the change of geometry
    for link_id, link in link_dict.items():
        link.lanes_change_point_list = [item / link.length * link.length_offset for item in link.lanes_change_point_list]


def _linkLengthToCut(link):
    # should not change number of lanes at the two ends of a macro link
    upstream_max_cut = max(_length_of_short_cut, link.lanes_change_point_list[1] - link.lanes_change_point_list[0] - 3)
    downstream_max_cut = max(_length_of_short_cut, link.lanes_change_point_list[-1] - link.lanes_change_point_list[-2] - 3)

    if link.upstream_short_cut and link.downstream_short_cut:
        total_length_needed = 2 * _length_of_short_cut + _minimum_cutted_length
        if link.length_offset > total_length_needed:
            link.length_of_cut_upstream = _length_of_short_cut
            link.length_of_cut_downstream = _length_of_short_cut
        else:
            link.length_of_cut_upstream = link.length_offset / total_length_needed * _length_of_short_cut
            link.length_of_cut_downstream = link.length_offset / total_length_needed * _length_of_short_cut

    elif link.upstream_short_cut:
        length_found = False
        ii = 0
        for i in range(link.lanes_list[-1], -1, -1):
            if link.length_offset > min(downstream_max_cut, _length_of_cut[i]) + _length_of_short_cut + _minimum_cutted_length:
                ii = i
                length_found = True
                break

        if length_found:
            link.length_of_cut_upstream = _length_of_short_cut
            link.length_of_cut_downstream = min(downstream_max_cut, _length_of_cut[ii])
        else:
            downstream_needed = min(downstream_max_cut, _length_of_cut[0])
            total_length_needed = downstream_needed + _length_of_short_cut + _minimum_cutted_length
            link.length_of_cut_upstream = link.length_offset / total_length_needed * _length_of_short_cut
            link.length_of_cut_downstream = link.length_offset / total_length_needed * downstream_needed

    elif link.downstream_short_cut:
        length_found = False
        ii = 0
        for i in range(link.lanes_list[-1], -1, -1):
            if link.length_offset > min(upstream_max_cut, _length_of_cut[i]) + _length_of_short_cut + _minimum_cutted_length:
                ii = i
                length_found = True
                break

        if length_found:
            link.length_of_cut_upstream = min(upstream_max_cut, _length_of_cut[ii])
            link.length_of_cut_downstream = _length_of_short_cut
        else:
            upstream_needed = min(upstream_max_cut, _length_of_cut[0])
            total_length_needed = upstream_needed + _length_of_short_cut + _minimum_cutted_length
            link.length_of_cut_upstream = link.length_offset / total_length_needed * _length_of_cut[0]
            link.length_of_cut_downstream = link.length_offset / total_length_needed * _length_of_short_cut

    else:
        length_found = False
        ii = 0
        for i in range(link.lanes_list[-1], -1, -1):
            if link.length_offset > min(upstream_max_cut, _length_of_cut[i]) + min(downstream_max_cut, _length_of_cut[i]) + _minimum_cutted_length:
                ii = i
                length_found = True
                break

        if length_found:
            link.length_of_cut_upstream = min(upstream_max_cut, _length_of_cut[ii])
            link.length_of_cut_downstream = min(downstream_max_cut, _length_of_cut[ii])
        else:
            upstream_needed = min(upstream_max_cut, _length_of_cut[0])
            downstream_needed = min(downstream_max_cut, _length_of_cut[0])
            total_length_needed = downstream_needed + upstream_needed + _minimum_cutted_length
            link.length_of_cut_upstream = link.length_offset / total_length_needed * upstream_needed
            link.length_of_cut_downstream = link.length_offset / total_length_needed * downstream_needed


def _performLinkCut(link, GT):
    link.cutted_lanes_change_point_list = link.lanes_change_point_list.copy()
    link.cutted_lanes_list = link.lanes_list.copy()
    link.cutted_lanes_change_list = link.lanes_change_list.copy()

    link.cutted_lanes_change_point_list[0], link.cutted_lanes_change_point_list[-1] = link.length_of_cut_upstream, link.length_offset - link.length_of_cut_downstream

    for i in range(len(link.cutted_lanes_list)):
        start_position = link.cutted_lanes_change_point_list[i]
        end_position = link.cutted_lanes_change_point_list[i+1]

        segment_geometry_xy = substring(link.geometry_xy_offset, start_dist=start_position, end_dist=end_position)
        link.cutted_geometry_xy_list.append(segment_geometry_xy)
        link.cutted_geometry_list.append(GT.geo_to_latlon(segment_geometry_xy))


def _cutMacroLinks(link_dict, GT):
    # cut macro links to stop lines to enable movement links generation
    for link_id, link in link_dict.items():
        _linkLengthToCut(link)
        _performLinkCut(link, GT)


def buildMultiResolutionNets(macronet, generate_micro_net=True,
                             auto_movement_generation=True, exclusive_bike_walk_lanes=True, connector_type=None,
                             width_of_lane=3.5, length_of_cell=7.0):
    """
    Build the corresponding mesoscopic and microscopic networks for a source (macroscopic) network

    Parameters
    ----------
    macronet: Network
        a soucrce osm2gmns Network object
    generate_micro_net: bool
        True: generate meso and micro networks; False: only generate meso network
    auto_movement_generation: bool
        automatically generate movements for intersections without movement information by calling function generateMovements
        in osm2gmns. if auto_movement_generation is set as False, movements at intersections without movement information will not be generated
    exclusive_bike_walk_lanes: bool
        build exclusive lanes for bike and walk
    connector_type: int
        link_type of connectors
    width_of_lane: float
        width of lanes in meter
    length_of_cell: float
        lenght of cells in meter

    Returns
    -------
    None

    """
    if og_settings.verbose:
        print('Building Multiresolution Networks')

    if not macronet.complete_link_lane_info:
        print('WARNING: Multiresolution network generation will be skipped because some links do not have lanes information.\n'
              '         If the network is parsed from a OSM file, set "default_lanes" as "True" when parsing networks to assign default lanes for links without lanes information.\n'
              '         If the network is loaded from a CSV file, make sure all links have lanes information before loading the network.')
        return


    if auto_movement_generation:
        generateMovements(macronet)
    else:
        for _, link in macronet.link_dict.items():
            link.linkLaneListFromSegment()
        validateUserInputMovements(macronet)

    _offsetLinkGeometry(macronet.link_dict, width_of_lane, macronet.GT)
    _checkMovementLinkNecessity(macronet.node_dict)
    _cutMacroLinks(macronet.link_dict, macronet.GT)

    net_generator = NetGenerator(macronet, generate_micro_net, exclusive_bike_walk_lanes, length_of_cell, width_of_lane)
    net_generator.generateNet()
    macronet.mesonet, macronet.micronet = net_generator.mesonet, net_generator.micronet



# -*- coding:utf-8 -*-
##############################################################
# Created Date: Monday, May 29th 2023
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################


import requests
import json
import urllib
import os
from typing import Union
from pathlib import Path


class OSMRelationIDFinder:
    """A class to find the osm relation id of a place of interest globally
       Use the Nominatim API to find the relation id of a place of interest

    Args:
        poi_name (str): the name of the place of interest,
            e.g., "Arizona State University"
            e.g., "Tempe, Arizona, USA"
            e.g., "Arizona, USA"

    Example:
    >>> import osm2gmns as og
    >>> rel = og.OSM_RelationID_Finder("Arizona State University")
    >>> rel.rel_id
        Info: Found relation id 3444656 from web
        Info: location of the place of interest:
        {
            "place_id": 318528634,
            "licence": "Data \u00a9 OpenStreetMap contributors, ODbL 1.0. http://osm.org/copyright",
            "osm_type": "relation",
            "osm_id": 3444656,
            "lat": "33.4213174",
            "lon": "-111.93316305413154",
            "class": "amenity",
            "type": "university",
            "place_rank": 30,
            "importance": 0.5547365758311374,
            "addresstype": "amenity",
            "name": "Arizona State University",
            "display_name": "Arizona State University, 1151, South Forest Avenue, Tempe Junction, Tempe, Maricopa County, Arizona, 85281, United States",
            "boundingbox": [
                "33.4102062",
                "33.4329786",
                "-111.9411651",
                "-111.9092447"
            ]
        }
    3444656

    # download the corresponding osm file
    >>> og.downloadOSMData(rel.rel_id, 'asu.osm')

    """

    def __init__(self, poi_name: str) -> None:

        self.poi_name = poi_name
        self.__url_json = self.__get_url_data()

        # maximum number of results to show
        # if parameter will applied in function rel_id_info
        # if the number of results is greater than _max_show, only show the first _max_show results
        self._max_show = 5

        # TODO: Unable to find the relation id locally, search from the web
        # the default path to save/load global country, state, and city information
        self._path_global_rel_id = self.__path2linux(os.path.join(
            Path(__file__).resolve().parent.parent.parent, "dependencies/global_rel_id.json"))

    def __get_url_data(self) -> dict:
        # prepare url
        base_url = 'https://nominatim.openstreetmap.org/search?'
        params = {"q": self.poi_name, "format": "json"}
        url = base_url + urllib.parse.urlencode(params)

        # get data from url using requests
        url_res = requests.get(url)
        url_json = json.loads(url_res.text)
        return url_json

    @property
    def rel_id(self) -> Union[float, None]:
        # Three logic to find the relation id
        # 1. Search from web, if relation id exist, return the relation id
        # 2. If relation id does not exist
        #    check whether url_json (result from web) is empty,
        #    if not, return relation id of the first element
        # 3. If url_json is empty, Search locally, return None if not found locally

        for i in self.__url_json:
            if i["osm_type"] == "relation":
                print(f"  Info: Found relation id {i['osm_id']} from web")
                print("  Info: location of the place of interest:")
                print(json.dumps(i, indent=4))
                return i["osm_id"]

        if len(self.rel_id_info) > 0:
            print("  Info: Could not find relation id directly from web")
            print(f"  but there is a possible relation id {self.rel_id_info[0]['osm_id']} from web")
            print("  or you can find more related ids: rel.rel_id_info")
            print(f"Detailed info for {self.rel_id_info[0]['osm_id']}: {json.dumps(self.rel_id_info[0], indent=4)}\n")
            return self.rel_id_info[0]["osm_id"]

        print("  Info: Could not find relation id from web: https://nominatim.openstreetmap.org/ui/search.html \n")
        return self.__find_local_rel_id()

    @property
    def rel_id_info(self) -> dict:
        # show detailed information of the relation id
        return self.__url_json if len(self.__url_json) < self._max_show else self.__url_json[:self._max_show]

    # convert OS path to standard linux path
    def __path2linux(self, path: Union[str, Path]) -> str:
        """Convert a path to a linux path, linux path can run in windows, linux and mac"""
        try:
            return path.replace("\\", "/")
        except Exception:
            return str(path).replace("\\", "/")

    def __find_local_rel_id(self) -> str:
        # sourcery skip: extract-duplicate-method

        # Step 0 load the global country, state, and city information from github
        try:
            print("  Info: Try to load global_rel_id.json from GitHub\n")
            github_url = "https://raw.githubusercontent.com/xyluo25/OSM2GMNS/master/dependencies/global_rel_id.json"
            global_rel_id_dict = json.loads(requests.get(github_url).text)
        except Exception:
            # if user can not access the internet, load the local file instead
            # please make sure the file exists in the local path
            # path: dependencies/global_rel_id.json
            print("  Info: Can not load global_rel_id.json from GitHub")
            print(f"  try to load local file: {self._path_global_rel_id}.\n")
            try:
                with open(self._path_global_rel_id, "r") as f:
                    global_rel_id_dict = json.load(f)
            except Exception:
                print("  Error: Can not load global_rel_id.json from local file\n \
                      please make sure the file exists in dependencies/global_rel_id.json \n \
                      Please search from https://www.openstreetmap.org/#map=5/40.298/-102.500  \n")
                return None

        # if we can not load dictionary, return None
        if not isinstance(global_rel_id_dict, dict):
            print("  Info: Could not load global_rel_id.json from GitHub.")
            print("  Please search from https://www.openstreetmap.org/#map=5/40.298/-102.500  \n")
            return None

        search_name = self.poi_name.split(",")[0].strip()
        if search_name == self.poi_name:
            search_name = self.poi_name.split(" ")[0].strip()

        # Step 1 Find the country start with the input characters
        start_with_country = [country_name for country_name in global_rel_id_dict["country"].keys(
        ) if country_name.lower().startswith(search_name.lower())][:self._max_show]

        # Step 2 Find the state start with the input characters
        start_with_state = [state_name for state_name in global_rel_id_dict["state"].keys(
        ) if state_name.lower().startswith(search_name.lower())][:self._max_show]

        # Step 3 Find the city start with the input characters
        start_with_city = [city_name for city_name in global_rel_id_dict["city"].keys(
        ) if city_name.lower().startswith(search_name.lower())][:self._max_show]

        # Step 4 Format the results
        formatted_str = ""
        if start_with_city:
            formatted_str += "City:\n"
            for city_name in start_with_city:
                formatted_str += f"""{city_name},
                {global_rel_id_dict["city"][city_name]["state"]},
                {global_rel_id_dict["city"][city_name]["country"]},
                rel_id: {global_rel_id_dict["city"][city_name]["osm_id"]} \n"""

        if start_with_state:
            formatted_str += "State:\n"
            for state_name in start_with_state:
                formatted_str += f"""{state_name},
                {global_rel_id_dict["state"][state_name]["country"]},
                rel_id: {global_rel_id_dict["state"][state_name]["osm_id"]} \n"""

        if start_with_country:
            formatted_str += "Country:\n"
            for country_name in start_with_country:
                formatted_str += f"""{country_name},
                rel_id: {global_rel_id_dict["country"][country_name]["osm_id"]} \n"""

        if formatted_str:

            formatted_str = f"""Info:\nWe can not search rel_id from web with place of interest: {self.poi_name} \n \
            Search locally with country, state and city startswith {search_name}:\n""" + formatted_str

            formatted_str += "You can replace your place of interest with local results above to get the rel_id\n"
            print(formatted_str)

        print("  Info: Could not load global_rel_id.json locally")
        print("  Please search from https://www.openstreetmap.org/#map=5/40.298/-102.500  \n")
        return None


def getOSMRelationID(poi_name: str) -> Union[float, None]:
    """Get the relation id of a place of interest globally

    Args:
        poi_name (str): the name of the place of interest,
            e.g., "Arizona State University"
            e.g., "Tempe, Arizona, USA"
            e.g., "Arizona, USA"

    Returns:
        Union[float, None]: the relation id of the place of interest, if not found, return None

    Example:
        >>> import osm2gmns as og
        >>> rel = og.getOSMRelationID("Arizona State University")
        >>> rel
            Info: Found relation id 3444656 from web
            Info: location of the place of interest:
            {
                "place_id": 318528634,
                "licence": "Data \u00a9 OpenStreetMap contributors, ODbL 1.0. http://osm.org/copyright",
                "osm_type": "relation",
                "osm_id": 3444656,
                "lat": "33.4213174",
                "lon": "-111.93316305413154",
                "class": "amenity",
                "type": "university",
                "place_rank": 30,
                "importance": 0.5547365758311374,
                "addresstype": "amenity",
                "name": "Arizona State University",
                "display_name": "Arizona State University, 1151, South Forest Avenue, Tempe Junction, Tempe, Maricopa County, Arizona, 85281, United States",
                "boundingbox": [
                    "33.4102062",
                    "33.4329786",
                    "-111.9411651",
                    "-111.9092447"
                ]
            }
        3444656
    """
    rel = OSMRelationIDFinder(poi_name)
    return rel.rel_id


from osm2gmns.osmnet.build_net import getNetFromFile
from osm2gmns.osmnet.combine_links import combineShortLinks
from osm2gmns.osmnet.complex_intersection import consolidateComplexIntersections
from osm2gmns.osmnet.enrich_net_info import generateNodeActivityInfo, generateLinkVDFInfo
from osm2gmns.osmnet.pois import connectPOIWithNet
from osm2gmns.osmnet.visualization import show, saveFig
from osm2gmns.movement.generate_movements import generateMovements
from osm2gmns.multiresolutionnet.build_mrnet import buildMultiResolutionNets
from osm2gmns.io.load_from_csv import loadNetFromCSV
from osm2gmns.io.downloader import downloadOSMData
from osm2gmns.io.writefile import outputNetToCSV
from osm2gmns.utils.util import config
from osm2gmns.func_lib.extract_relation_id import getOSMRelationID
import osm2gmns.settings as og_settings


__version__ = '0.7.6'
print(f'osm2gmns, {__version__}')

__all__ = [
    'getNetFromFile',
    'combineShortLinks',
    'consolidateComplexIntersections',
    'generateNodeActivityInfo',
    'generateLinkVDFInfo',
    'connectPOIWithNet',
    'show',
    'saveFig',
    'generateMovements',
    'buildMultiResolutionNets',
    'loadNetFromCSV',
    'downloadOSMData',
    'outputNetToCSV',
    'config',
    'getOSMRelationID',
    'og_settings'
]



# todo: railway allowed_use

# todo: turns
# todo: parse speed

# todo: link_id: int, str
# todo: compatible with networkx
# todo: mrnet bike walk
# todo: movement generation

# todo: default speed railway

# todo: og.link_types
# todo: combine_link remove segment

# TODO: find global relation id locally (user can update locally as well)


# -*- coding:utf-8 -*-
##############################################################
# Created Date: Tuesday, May 30th 2023
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################

# Please note,
# All countries, states, and cities in global_rel_id.json file are from https://github.com/dr5hn/countries-states-cities-database/tree/master
from .extract_relation_id import getOSMRelationID

__all__ = ['getOSMRelationID']


# -*- coding:utf-8 -*-
##############################################################
# Created Date: Sunday, February 4th 2024
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################


from osm2gmns.func_lib.extract_relation_id import OSMRelationIDFinder


def test_rel_id_web_search_exists():

    rel_identifier = OSMRelationIDFinder("Arizona State University")
    rel_id = rel_identifier.rel_id
    print(type(rel_id))
    assert isinstance(rel_id, int)


def test_rel_id_web_search_not_exists():

    rel_identifier = OSMRelationIDFinder("TTTTTTEEEEEESSSSSTTTTT")
    rel_id = rel_identifier.rel_id
    assert rel_id is None



import setuptools

try:
    # if have requirements.txt file inside the folder
    with open("requirements.txt", "r", encoding="utf-8") as f:
        modules_needed = [i.strip() for i in f.readlines()]
except Exception:
    modules_needed = []

with open("README.rst", "r", encoding="utf-8") as f:
    _long_description = f.read()

setuptools.setup(
    name='osm2gmns',
    version='0.7.6',
    author='Jiawei Lu, Xuesong Zhou',
    author_email='jiaweil9@asu.edu, xzhou74@asu.edu',
    url='https://github.com/jiawlu/OSM2GMNS',
    description="convert map data from OpenStreetMap to network files in GMNS format",
    long_description=_long_description,
    long_description_content_type="text/x-rst",
    license='GPLv3+',
    packages=setuptools.find_packages(),
    python_requires=">=3.10",
    classifiers=['License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.10',
                 'Programming Language :: Python :: 3.11',
                 'Programming Language :: Python :: 3.12'
                 ],
    install_requires=modules_needed,
)
